# -*- coding: utf-8 -*-
"""
pyqdap.py

Module based on R qdap package:
http://cran.r-project.org/web/packages/qdap/index.html

Very limited functionality as-yet.

Created on Fri Jun 19 16:17:38 2015

@author: mthornton
"""

import csv
import re
import string
import itertools
import nltk

# read in lists of words
wordlists = {'positive': [], 'negative': [], 'negation': [],
             'amplifier': [], 'deamplifier': [], 'swear': [], 'acronyms': []}
wordlistkeys = wordlists.keys()
for k in wordlistkeys:
    with open(k + '.csv', 'rb') as dfile:
        reader = csv.reader(dfile)
        wordlists[k] = list(itertools.chain(*reader))
wordweights = {}
try:
    with open('wordweights.csv', 'rb') as dfile:
        reader = csv.reader(dfile)
        for row in reader:
            wordweights[row[0]] = float(row[1])
except:
    for w in wordlists['positive']:
        wordweights[w] = 1.
    for w in wordlists['negative']:
        wordweights[w] = -1.


class corpus:
    """Primary class of this module - feed in text and do things with it!"""

    def __init__(self, rawtext):
        self.rawtext = rawtext
        self.shiftag = False
        self.poltag = False
        self.tokenize()

    def tokenize(self):
        """Break text into sentences and words for later processing"""
        self.sentences = nltk.sent_tokenize(self.rawtext)
        self.nsent = len(self.sentences)
        self.sentwords = [nltk.word_tokenize(s) for s in self.sentences]
        self.lowsentwords = []
        for s in self.sentwords:
            self.lowsentwords.append([])
            for w in s:
                self.lowsentwords[-1].append(w.lower())
        self.allwordsnpunct = list(itertools.chain(*self.sentwords))
        self.allwords = []
        for w in self.allwordsnpunct:
            wt = re.sub(r'[^a-zA-Z\s]', '', w)
            if not wt == '':
                self.allwords.append(wt)
        self.wordn = len(self.allwords)

    def polarity(self, questweight=0, ampweight=0.8, nbefore=4, nafter=2):
        """Assess polarity (i.e. valence) of corpus"""
        if hasattr(self, 'pole'):
            return(self.pole)
        if self.wordn == 0:
            self.pole = 0.
            return(self.pole)
        if self.poltag:
            pol = self.taggedpol
        else:
            pol = [self.tag_polarized(s) for s in self.lowsentwords]
            self.taggedpol = pol
            self.poltag = True
        if self.shiftag:
            shift = self.taggedshift
        else:
            shift = [self.tag_shifters(s) for s in self.lowsentwords]
            self.taggedshift = shift
            self.shiftag = True
        polsum = []
        for i, p in enumerate(pol):
            polsum.append(self.context(p, shift[i], questweight, ampweight,
                                       nbefore, nafter))
        # to account for the fact that redoubled amps can make abs(pol) > 1
        self.pole = sorted([-1, sum(polsum)/self.wordn, 1])[1]
        return(self.pole)

    def tag_polarized(self, sentlist):
        """Tags the position of polarized words in sentence"""
        tagged = []
        for w in sentlist:
            if (w in wordlists['positive']) or (w in wordlists['negative']):
                tagged.append(wordweights[w])
            else:
                tagged.append(0.)
        return(tagged)

    def tag_shifters(self, sentlist):
        """Tags the position of shifter words in sentence"""
        tagged = []
        for w in sentlist:
            if w in wordlists['negation']:
                tagged.append('n')
            elif w in wordlists['amplifier']:
                tagged.append('a')
            elif w in wordlists['deamplifier']:
                tagged.append('d')
            elif w == ',':
                tagged.append('c')
            elif w == '!':
                tagged.append('e')
            elif w == '?':
                tagged.append('q')
            elif w in string.punctuation:
                tagged.append('p')
            else:
                tagged.append('')
        return(tagged)

    def context(self, pol, shift, qwei, awei, nbe, naf):
        """Modifies polarity based on shifters in context"""
        if qwei == 0:
            if shift[-1] == 'q':
                return(0.)
        polmod = []
        for i, p in enumerate(pol):
            if not p == 0:
                bef = 1
                damp = 0
                neg = 0
                lim = 0
                while bef < (nbe + 1 + lim):
                    ind = i - bef
                    if ind < 0:
                        break
                    elif shift[ind] == 'n':
                        neg += 1
                    elif shift[ind] == 'a':
                        damp += 1.
                    elif shift[ind] == 'd':
                        damp -= 1.
                    elif shift[ind] == 'c':
                        break
                    elif shift[ind] in ['e', 'p', 'q']:
                        lim += 1
                    bef += 1
                aft = 1
                lim = 0
                nws = len(pol)
                while aft < naf + 1 + lim:
                    ind = i + aft
                    if ind >= nws:
                        break
                    elif shift[ind] == 'n':
                        neg += 1
                    elif shift[ind] == 'a':
                        damp += 1.
                    elif shift[ind] == 'd':
                        damp -= 1.
                    elif shift[ind] == 'c':
                        break
                    elif shift[ind] in ['e', 'p', 'q']:
                        lim += 1
                    aft += 1
                pm = p + p * awei * damp
                if shift[-1] == 'q':
                    pm = pm*qwei
                if neg % 2:
                    pm = pm * -1
                polmod.append(pm)
        return(sum(polmod))

    def intensity(self, measure, filteracronyms=True):
        """Assess polarity (i.e. valence) of corpus"""
        if self.wordn == 0:
            self.arousal = 0.
            return(self.arousal)
        if measure == 'abspol':
            arousal = self.abspol()/self.wordn
        elif measure == "shiftdiff":
            arousal = self.shiftdiff()/self.wordn/2.+.5
        elif measure == "exclamation":
            arousal = self.exclacount()/self.nsent
        elif measure == "caps":
            arousal = self.capscount(filteracronyms)/self.wordn
        elif measure == 'swear':
            arousal = self.profane()/self.wordn
        elif measure == 'composite':
            abspol = self.abspol()/self.wordn
            profane = self.profane()/self.wordn
            shiftdiff = self.shiftdiff()/self.wordn/2.+.5
            exclamation = self.exclacount()/self.nsent
            # weights from calibration sds
            #arousal = -.262*pol/.118 + .457*abspol/.11 + .748*profane/.029
            # weights from initial time series
            arousal = (.567*abspol/.111 + .142*shiftdiff/.050 + .094*exclamation/.211 + .678*profane/.028)/4
        else:
            raise ValueError('Non-accepted measure or no measure passed')
        self.arousal = arousal
        return(arousal)

    def abspol(self):
        """Calculates the number of polaized words"""
        if self.poltag:
            pol = self.taggedpol
        else:
            pol = [self.tag_polarized(s) for s in self.lowsentwords]
        npol = 0.
        for p in pol:
            if not type(p) == list:
                p = [p, 0]
            for w in p:
                if not w == 0:
                    npol += 1
        self.npol = npol
        return(npol)

    def shiftdiff(self):
        """Calculates differences im (de)amplifiers"""
        if self.shiftag:
            shift = self.taggedshift
        else:
            shift = [self.tag_shifters(s) for s in self.lowsentwords]
        ampdif = 0.
        for s in shift:
            if not type(s) == list:
                s = [s, '']
            for w in s:
                if w == 'a':
                    ampdif += 1
                elif w == 'd':
                    ampdif -= 1
        self.ampdif = ampdif
        return(ampdif)

    def exclacount(self):
        """Calculate the number of sentence terminating exclamations"""
        excount = 0.
        for s in self.sentwords:
            if s[-1] == "!":
                excount += 1
        self.excount = excount
        return(excount)

    def capscount(self, acfilt):
        """Counts all caps words"""
        capn = 0.
        for s in self.sentwords:
            for w in s:
                if w.isupper() and len(w) > 3:
                    if acfilt:
                        if not w in wordlists['acronyms']:
                            capn += 1
                    else:
                        capn += 1
        self.capn = capn
        return(capn)

    def profane(self):
        """Count all swear words"""
        cursecount = 0.
        for s in self.sentwords:
            for w in s:
                if w in wordlists['swear']:
                    cursecount += 1
        self.cursecount = cursecount
        return(cursecount)
